#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
gen.py: Pre-arranged AETVgen usage
"""

__author__     = "Ekawat (Ice) Homsirikamol"
__copyright__  = "Copyright (c) 2015 CERG @ GMU, USA"
__license__    = "GPL 3.0"
__version__    = "1.0b1"

# Core imports
import random
import sys
import AETVgen
import shutil

repo = "../../../../"
copy = False
    
    
def AES_GCM(size, testMode, verbose, decrypt):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSizeM': 1, 'maxSizeM': 64,
            'minSizeAD': 1, 'maxSizeAD': 64,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt,
            'maxSizeSegment' : 10000}
    aetv = AETVgen.AETVgen('aes-gcm', **args)
    aetv.genTestVectors(size)
    # Copy generated files to hardware repository
    if copy == True:
        dst = repo + "AES_GCM/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "AES_GCM/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)


def ICEPOLE(size, testMode, verbose, decrypt):
    args = {'stdout': False, 'sizePIO': 32, 'sizeSIO': 4, 'enableNSEC': False,
            'minSizeM': 100, 'maxSizeM': 300,
            'minSizeAD': 1, 'maxSizeAD': 128,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt,
            'maxSizeSegment' : 256}
    aetv = AETVgen.AETVgen('ICEPOLE', **args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "ICEPOLE/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)
        dst = repo + "ICEPOLE/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)


def Minalpher(size, testMode, verbose, decrypt):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNPUB': 13,
            'minSizeM': 1, 'maxSizeM': 1000,
            'minSizeAD': 1, 'maxSizeAD': 1000,
            'verbose': verbose, 'testMode': testMode,'decrypt': decrypt,
            'maxSizeSegment' : 64}
    aetv = AETVgen.AETVgen('minalpher', **args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Minalpher/x1/HLS/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Joltik(size, testMode, verbose, decrypt):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'minSizeM': 1, 'maxSizeM': 16,
            'sizeTag': 8, 'sizeKey': 16, 'sizeNPUB': 4,
            'minSizeAD': 1, 'maxSizeAD': 16,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt,
            'maxSizeSegment' : 32}
    aetv = AETVgen.AETVgen('Joltik', **args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Joltik/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)

def Scream(size, testMode, verbose, decrypt):
    args = {'stdout': False, 'sizePIO': 4, 'sizeSIO': 4,
            'sizeTag': 16, 'sizeKey': 16, 'sizeNPUB': 12,
            'minSizeM': 1, 'maxSizeM': 32,
            'minSizeAD': 1, 'maxSizeAD': 32,
            'verbose': verbose, 'testMode': testMode, 'decrypt': decrypt,
            'maxSizeSegment' : 64}
    aetv = AETVgen.AETVgen('scream', **args)
    aetv.genTestVectors(size)
    # Copy files to hardware repository
    if copy == True:
        dst = repo + "Joltik/x1/RTL/scripts/"
        for file in ["do.txt", "pdi.txt", "sdi.txt"]:
            shutil.copyfile(file, dst + file)    

if __name__ == '__main__':
    testMode = 1
    verbose  = False
    decrypt  = False
    AES_GCM(20, testMode, verbose, decrypt)
    #ICEPOLE(20, testMode, verbose, decrypt)
    #Joltik(20, testMode, verbose, decrypt)
    #Minalpher(20, testMode, verbose, decrypt)
    #Scream(20, testMode, verbose, decrypt)




